@@ -7,8 +7,12 @@
 #
 # Copyright (c) 2018  Paul T. McGuire
 #
+from __future__ import division
 
-import unittest
+try:
+    import unittest2 as unittest
+except ImportError:
+    import unittest
 import pyparsing as pp
 from collections import namedtuple
 from datetime import datetime
@@ -25,6 +29,16 @@ class PyparsingExpressionTestCase(unittest.TestCase):
     given text strings. Subclasses must define a class attribute 'tests' which
     is a list of PpTestSpec instances.
     """
+    
+    if not hasattr(unittest.TestCase, 'subTest'):
+        # Python 2 compatibility
+        from contextlib import contextmanager
+        @contextmanager
+        def subTest(self, **params):
+            print('subTest:', params)
+            yield
+    
+    tests = []
     def runTest(self):
         if self.__class__ is PyparsingExpressionTestCase:
             return
@@ -32,16 +46,16 @@ class PyparsingExpressionTestCase(unittest.TestCase):
         for test_spec in self.tests:
             # for each spec in the class's tests list, create a subtest
             # that will either:
-            #  - parse the string with expected success, display the 
+            #  - parse the string with expected success, display the
             #    results, and validate the returned ParseResults
-            #  - or parse the string with expected failure, display the 
+            #  - or parse the string with expected failure, display the
             #    error message and mark the error location, and validate
             #    the location against an expected value
             with self.subTest(test_spec=test_spec):
                 test_spec.expr.streamline()
-                print("\n{} - {}({})".format(test_spec.desc, 
-                                             type(test_spec.expr).__name__, 
-                                             test_spec.expr))
+                print("\n{0} - {1}({2})".format(test_spec.desc,
+                                                type(test_spec.expr).__name__,
+                                                test_spec.expr))
 
                 parsefn = getattr(test_spec.expr, test_spec.parse_fn)
                 if test_spec.expected_fail_locn is None:
@@ -64,18 +78,23 @@ class PyparsingExpressionTestCase(unittest.TestCase):
                         # compare results against given list and/or dict
                         if test_spec.expected_list is not None:
                             self.assertEqual([result], test_spec.expected_list)
-
                 else:
                     # expect fail
-                    with self.assertRaises(pp.ParseException) as ar:
+                    try:
                         parsefn(test_spec.text)
-                    print(' ', test_spec.text or "''")
-                    print(' ', ' '*ar.exception.loc+'^')
-                    print(' ', ar.exception.msg)
-                    self.assertEqual(ar.exception.loc, test_spec.expected_fail_locn)
+                    except Exception as exc:
+                        if not hasattr(exc, '__traceback__'):
+                            # Python 2 compatibility
+                            from sys import exc_info
+                            etype, value, traceback = exc_info()
+                            exc.__traceback__ = traceback
+                        print(pp.ParseException.explain(exc))
+                        self.assertEqual(exc.loc, test_spec.expected_fail_locn)
+                    else:
+                        self.assertTrue(False, "failed to raise expected exception")
 
 
-#=========== TEST DEFINITIONS START HERE ==============
+# =========== TEST DEFINITIONS START HERE ==============
 
 class TestLiteral(PyparsingExpressionTestCase):
     tests = [
@@ -149,6 +168,22 @@ class TestWord(PyparsingExpressionTestCase):
         ),
     ]
 
+class TestCombine(PyparsingExpressionTestCase):
+    tests = [
+        PpTestSpec(
+            desc="Parsing real numbers - fail, parsed numbers are in pieces",
+            expr=pp.OneOrMore(pp.Word(pp.nums) + '.' + pp.Word(pp.nums)),
+            text="1.2 2.3 3.1416 98.6",
+            expected_list=['1', '.', '2', '2', '.', '3', '3', '.', '1416', '98', '.', '6'],
+        ),
+        PpTestSpec(
+            desc="Parsing real numbers - better, use Combine to combine multiple tokens into one",
+            expr=pp.OneOrMore(pp.Combine(pp.Word(pp.nums) + '.' + pp.Word(pp.nums))),
+            text="1.2 2.3 3.1416 98.6",
+            expected_list=['1.2', '2.3', '3.1416', '98.6'],
+        ),
+    ]
+
 class TestRepetition(PyparsingExpressionTestCase):
     tests = [
         PpTestSpec(
@@ -214,7 +249,7 @@ class TestGroups(PyparsingExpressionTestCase):
     tests = [
         PpTestSpec(
             desc = "Define multiple results names in groups",
-            expr = pp.OneOrMore(pp.Group(pp.Word(pp.alphas)("key") 
+            expr = pp.OneOrMore(pp.Group(pp.Word(pp.alphas)("key")
                                           + EQ
                                           + pp.pyparsing_common.number("value"))),
             text = "range=5280 long=-138.52 lat=46.91",
@@ -222,7 +257,7 @@ class TestGroups(PyparsingExpressionTestCase):
         ),
         PpTestSpec(
             desc = "Define multiple results names in groups - use Dict to define results names using parsed keys",
-            expr = pp.Dict(pp.OneOrMore(pp.Group(pp.Word(pp.alphas) 
+            expr = pp.Dict(pp.OneOrMore(pp.Group(pp.Word(pp.alphas)
                                           + EQ
                                           + pp.pyparsing_common.number))),
             text = "range=5280 long=-138.52 lat=46.91",
@@ -243,22 +278,28 @@ class TestGroups(PyparsingExpressionTestCase):
 
 class TestParseAction(PyparsingExpressionTestCase):
     tests = [
+        PpTestSpec(
+            desc="Parsing real numbers - use parse action to convert to float at parse time",
+            expr=pp.OneOrMore(pp.Combine(pp.Word(pp.nums) + '.' + pp.Word(pp.nums)).addParseAction(lambda t: float(t[0]))),
+            text="1.2 2.3 3.1416 98.6",
+            expected_list= [1.2, 2.3, 3.1416, 98.6], # note, these are now floats, not strs
+        ),
         PpTestSpec(
             desc = "Match with numeric string converted to int",
             expr = pp.Word("0123456789").addParseAction(lambda t: int(t[0])),
             text = "12345",
-            expected_list = [12345],  # note - result is type int, not str 
+            expected_list = [12345],  # note - result is type int, not str
         ),
         PpTestSpec(
             desc = "Use two parse actions to convert numeric string, then convert to datetime",
-            expr = pp.Word(pp.nums).addParseAction(lambda t: int(t[0]), 
+            expr = pp.Word(pp.nums).addParseAction(lambda t: int(t[0]),
                                                    lambda t: datetime.utcfromtimestamp(t[0])),
             text = "1537415628",
             expected_list = [datetime(2018, 9, 20, 3, 53, 48)],
         ),
         PpTestSpec(
             desc = "Use tokenMap for parse actions that operate on a single-length token",
-            expr = pp.Word(pp.nums).addParseAction(pp.tokenMap(int), 
+            expr = pp.Word(pp.nums).addParseAction(pp.tokenMap(int),
                                                    pp.tokenMap(datetime.utcfromtimestamp)),
             text = "1537415628",
             expected_list = [datetime(2018, 9, 20, 3, 53, 48)],
@@ -279,8 +320,9 @@ class TestParseAction(PyparsingExpressionTestCase):
 
 class TestResultsModifyingParseAction(PyparsingExpressionTestCase):
     def compute_stats_parse_action(t):
-        # by the time this parse action is called, parsed numeric words have been converted to ints
-        # by a previous parse action, so they can be treated as ints
+        # by the time this parse action is called, parsed numeric words
+        # have been converted to ints by a previous parse action, so
+        # they can be treated as ints
         t['sum'] = sum(t)
         t['ave'] = sum(t) / len(t)
         t['min'] = min(t)
@@ -296,6 +338,16 @@ class TestResultsModifyingParseAction(PyparsingExpressionTestCase):
         ),
     ]
 
+class TestRegex(PyparsingExpressionTestCase):
+    tests = [
+        PpTestSpec(
+            desc="Parsing real numbers - using Regex instead of Combine",
+            expr=pp.OneOrMore(pp.Regex(r'\d+\.\d+').addParseAction(lambda t: float(t[0]))),
+            text="1.2 2.3 3.1416 98.6",
+            expected_list=[1.2, 2.3, 3.1416, 98.6],  # note, these are now floats, not strs
+        ),
+    ]
+
 class TestParseCondition(PyparsingExpressionTestCase):
     tests = [
         PpTestSpec(
@@ -321,7 +373,7 @@ class TestTransformStringUsingParseActions(PyparsingExpressionTestCase):
     }
     def markup_convert(t):
         htmltag = TestTransformStringUsingParseActions.markup_convert_map[t.markup_symbol]
-        return "<{}>{}</{}>".format(htmltag, t.body, htmltag)
+        return "<{0}>{1}</{2}>".format(htmltag, t.body, htmltag)
 
     tests = [
         PpTestSpec(
@@ -392,24 +444,26 @@ class TestCommonHelperExpressions(PyparsingExpressionTestCase):
     ]
 
 
-#============ MAIN ================
+def _get_decl_line_no(cls):
+    import inspect
+    return inspect.getsourcelines(cls)[1]
+
+
+# get all test case classes defined in this module and sort them by decl line no
+test_case_classes = list(PyparsingExpressionTestCase.__subclasses__())
+test_case_classes.sort(key=_get_decl_line_no)
+
+# make into a suite and run it - this will run the tests in the same order
+# they are declared in this module
+#
+# runnable from setup.py using "python setup.py test -s simple_unit_tests.suite"
+#
+suite = unittest.TestSuite(cls() for cls in test_case_classes)
+
+
+# ============ MAIN ================
 
 if __name__ == '__main__':
-    # we use unittest features that are in Py3 only, bail out if run on Py2
-    import sys
-    if sys.version_info[0] < 3:
-        print("simple_unit_tests.py runs on Python 3 only")
-        sys.exit(0)
-        
-    import inspect
-    def get_decl_line_no(cls):
-        return inspect.getsourcelines(cls)[1]
+    result = unittest.TextTestRunner().run(suite)
 
-    # get all test case classes defined in this module and sort them by decl line no
-    test_case_classes = list(PyparsingExpressionTestCase.__subclasses__())
-    test_case_classes.sort(key=get_decl_line_no)
-    
-    # make into a suite and run it - this will run the tests in the same order
-    # they are declared in this module
-    suite = unittest.TestSuite(cls() for cls in test_case_classes)
-    unittest.TextTestRunner().run(suite)
+    exit(0 if result.wasSuccessful() else 1)
